/*
 * This progam extracts statistical parameters on chemical shifts
 * from BMRB files.
 *
 * Copyright (C) 2003 Antoine MARIN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1 published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact the author via mail or email at:
 * Antoine Marin (Antoine.Marin@ibpc.fr)
 * CNRS UPR 9080
 * Laboratoire de Biochimie Thorique
 * 13, rue Pierre et Marie Curie
 * 75005 Paris - France
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h> /* for basename */
#include <search.h> /* for hash tables */
#include <ctype.h>  /* for isdigit */
#include <values.h>
#include <math.h>
#include <limits.h> /* needed with IRIX 6.5 */
#include <float.h>  /* needed with IRIX 6.5 */

#include "alloc_utils.h"

#define VERSION "1.0"

#define MAX_HIST 5
#define MAX_BUF 1024
#define MAX_LONG_BUF 100000

#define MAX_NAME_LEN 8

#define EPSILON 1e-5

/*! \def MAX_CS_NUM
 * Maximum number of chemical shifts observed for a given atom name
 * (up to 10000 in the 2370 BMRB files in Nov. 2002)
 */
#define MAX_CS_NUM 20000

/*! \def MIN_CS_VAL
 * Minimum chemical shift value under which a warning is displayed
 * and the corresponding value is not taken into account
 */
#define MIN_CS_VAL -10.0

/*! \def MAX_CS_VAL
 * Maximum chemical shift value above which a warning is displayed
 * and the corresponding value is not taken into account
 */
#define MAX_CS_VAL 250.0

/*! \def DEF_VERB
 * Default verbosity level
 */
#define DEF_VERB 2

/*! \def DEF_IN_FILE
 * Default input file name
 */
#define DEF_IN_FILE "cs.in"

/*! \def DEF_OUT_FILE
 * Default output file name
 */
#define DEF_OUT_FILE "BMRB_stats.out"

/*! \def DEF_BMRB_DIR
 * Default BMRB directory
 */
#define DEF_BMRB_DIR "/home/share/BMRB"

/*! \def MIN_VERB
 * Minimum verbosity level
 */
#define MIN_VERB 0

/*! \def MAX_VERB
 * Maximum verbosity level
 */
#define MAX_VERB 3

char *AA3[] = {"ALA", "ARG", "ASP", "ASN", "CYS", "GLU", "GLN", "GLY", "HIS",
 "ILE", "LEU", "LYS", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL"
};

#undef DEBUG

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
typedef struct parameters {
  int verbosity;             /**< verbosity level */
  int print_cor_coefs;       /**< print correlation coefficients */
  int print_latex_cor_coefs; /**< print correlation corefficients in latex
                                  format */
  int group_atom_names;      /**< flag to group atom names
                                  (eg: HB1 + HB2 = HB) */
  char *in_fname;            /**< input file containing atoms to analyse */
  char *out_fname;           /**< output statistics file */
  char *BMRB_dir;            /**< input BMRB directory */
} Parameters;

typedef struct class_data {
  int size;         /**< matrix size */
  float **cov;      /**< covariances matrix */
  float **cor_coef; /**< correlations coefficients */
} Class_data;

typedef struct data {
  char **atom_desc;  /**< keyword for atom description of a chemical shift*/
  float **cs_val;    /**< chemical shift values for each atom type */
  int **cs_val_tag;  /**< 1 if the CS is present, 0 otherwise */
  float *cs_mean;    /**< chemical shifts means for each atom type */
  float *cs_stdev;   /**< chemical shifts standard deviation for each atom type */
  float *cs_min;     /**< chemical shifts minimums for each atom type */
  float *cs_max;     /**< chemical shifts maximums for each atom type */
  int *cs_count;     /**< number of chemical shifts for each atom type */
  int *cs_missing;   /**< missing chemical shifts for each atom type */
  int *cs_res;       /**< residue to which belong a chemical shift */
  int *res_count;    /**< number of each residue type encountered */
  int num_nuclei;    /**< number of different nuclei to analyse */
  Class_data *cdata; /**< data associated with each residue type */
} Data;

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void init_parameters(Parameters *params)
{
  params->verbosity = DEF_VERB;
  params->print_cor_coefs = 0;
  params->print_latex_cor_coefs = 0;
  params->group_atom_names = 0;
  params->in_fname = DEF_IN_FILE;
  params->out_fname = DEF_OUT_FILE;
  params->BMRB_dir = DEF_BMRB_DIR;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_usage(FILE *fp, char **argv)
{
  fprintf(fp, "\nUsage: %s [options] < BMRB_file_list.txt\n\n",
      basename(argv[0]));
  fprintf(fp, "  Options:\n\n");
  fprintf(fp, "    -bd directory : BMRB directory (default = %s).\n\n",
      DEF_BMRB_DIR);
  fprintf(fp, "    -gan          : group atom names (eg: HB1 + HB2 = HB)\n");
  fprintf(fp, "                    WARNING: this option can be used to study "
      "statistics\n                    but CANT'T be used for prediction.\n\n");
  fprintf(fp, "    -h, --help    : print this help.\n\n");
  fprintf(fp, "    -i file       : input file containing atoms to analyse "
      "(default = %s).\n", DEF_IN_FILE);
  fprintf(fp, "                    File format is:\n");
  fprintf(fp, "                    Amino   Atom    Atom\n");
  fprintf(fp, "                    Acid    Name    Type\n");
  fprintf(fp, "                    ALA     HA       H\n");
  fprintf(fp, "                    ALA     CB       C\n");
  fprintf(fp, "                    ...     ...      ...\n\n");
  fprintf(fp, "    -o file       : output statistics file "
      "(default = %s).\n\n", DEF_OUT_FILE);
  fprintf(fp, "    -pcc          : print correlation coefficients.\n\n");
  fprintf(fp, "    -plcc         : print correlation coefficients "
      "in LateX format.\n\n");
  fprintf(fp, "    -v, --version : print version only (current version = %s).\n\n",
      VERSION);
  fprintf(fp, "    -verb int     : verbosity level (default = %d).\n", DEF_VERB);
  fprintf(fp, "                    0 -> quiet\n");
  fprintf(fp, "                    1 -> errors\n");
  fprintf(fp, "                    2 -> errors + warnings\n");
  fprintf(fp, "                    3 -> errors + warnings + more "
      "informations\n\n");
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void process_command_line(int argc, char **argv, Parameters *params)
{
  int i;

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-bd") == 0) {

/* -bd : BMRB directory */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-bd' option.\n");
          print_usage(stderr, argv);
        }
        exit(1);
      }
      ALLOC(params->BMRB_dir, strlen(argv[i]) + 1, char);
      strcpy(params->BMRB_dir, argv[i]);
      if (params->BMRB_dir[strlen(params->BMRB_dir) - 1] == '/') {
        params->BMRB_dir[strlen(params->BMRB_dir) - 1] = '\0';
      }
    } else if (strcmp(argv[i], "-gan") == 0) {

/* -gan : group atom names */

      params->group_atom_names = 1;
      fprintf(stdout, "WARNING: grouping atom names is only usefull for "
          "studying\n         statistics, but CAN'T be used for prediction.\n");
    } else if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {

/* -h, --help : print the usage message */

      print_usage(stdout, argv);
      exit(0);
    } else if (strcmp(argv[i], "-i") == 0) {

/* -i : input file */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-i' option.\n");
          print_usage(stderr, argv);
        }
        exit(1);
      }
      ALLOC(params->in_fname, strlen(argv[i]) + 1, char);
      strcpy(params->in_fname, argv[i]);
    } else if (strcmp(argv[i], "-o") == 0) {

/* -o : output file */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-o' option.\n");
          print_usage(stderr, argv);
        }
        exit(1);
      }
      ALLOC(params->out_fname, strlen(argv[i]) + 1, char);
      strcpy(params->out_fname, argv[i]);
    } else if (strcmp(argv[i], "-pcc") == 0) {

/* -pcc : print correlation coefficients */

      params->print_cor_coefs = 1;
    } else if (strcmp(argv[i], "-plcc") == 0) {

/* -plcc : print correlation coefficients in LateX format */

      params->print_latex_cor_coefs = 1;
    } else if (strcmp(argv[i], "-v") == 0 ||
      strcmp(argv[i], "--version") == 0) {

/* -v, --version : print version and exit */

      fprintf(stdout, "This is %s release number %s\n", basename(argv[0]),
          VERSION);
      exit(0);
    } else if (strcmp(argv[i], "-verb") == 0) {

/* -verb : verbosity */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-v' option.\n");
          print_usage(stderr, argv);
        }
        exit(1);
      }
      if (isdigit((int) argv[i][0]) == 0) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: the value for the '-v' option "
              "should be a positive digit [%d-%d] (you provided: %s).\n",
              MIN_VERB, MAX_VERB, argv[i]);
          print_usage(stderr, argv);
        }
        exit(1);
      }
      params->verbosity = atoi(argv[i]);
      if (params->verbosity < MIN_VERB || params->verbosity > MAX_VERB) {
        fprintf(stderr, "Error: verbosity level out of range: %d.\n",
            params->verbosity);
        print_usage(stderr, argv);
        exit(1);
      }
    } else {

/* Unknown tag */

      if (params->verbosity >= 1) {
        fprintf(stderr, "Error: unknown tag: '%s'.\n", argv[i]);
        print_usage(stderr, argv);
      }
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void check_parameters(Parameters *params, char **argv)
{
  if (params->BMRB_dir[0] == '\0') {
    fprintf(stderr, "Error: BMRB directory is mandatory.\n");
    print_usage(stderr, argv);
    exit(1);
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void get_atoms_to_analyse(Parameters *params, Data *data)
{
  char buffer[MAX_LONG_BUF], amino_acid[MAX_NAME_LEN], atom_name[MAX_NAME_LEN];
  char atom_type[MAX_NAME_LEN];
  int lines;
  FILE *in_fp;

  FOPEN(in_fp, params->in_fname, "r");

  lines = 0;
  while (fgets(buffer, MAX_LONG_BUF, in_fp) != NULL) {
    if (buffer[0] != '#' &&
        sscanf(buffer, "%s%s%s", amino_acid, atom_name, atom_type) == 3) {
      lines++;
    }
  }
  if (lines == 0) {
    fprintf(stderr, "Error: there is no atom types to analyse in "
        "file '%s'.\n", params->in_fname);
    exit(1);
  }

  data->num_nuclei = lines;
  ALLOC2(data->atom_desc, lines, 16, char);

  rewind(in_fp);
  lines = 0;
  while (fgets(buffer, MAX_LONG_BUF, in_fp) != NULL) {
    if (buffer[0] != '#' &&
        sscanf(buffer, "%s%s%s", amino_acid, atom_name, atom_type) == 3) {
      if (params->group_atom_names == 0) {
        sprintf(data->atom_desc[lines], "%s %s %c", amino_acid, atom_name,
            atom_type[0]);
      } else {
        sprintf(data->atom_desc[lines], "%s %.2s %c", amino_acid, atom_name,
            atom_type[0]);
      }
      lines++;
    }
  }

  FCLOSE(in_fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Initialize hash table for full types (amino-acid + atom name + atom type)
 */
void init_htbl(Data *data, Parameters *params)
{
  int i, j;
  char amino_acid[16];
  ENTRY ent, *ep;

  ALLOC(data->cs_res, data->num_nuclei, int);

  for (i = 0; i < data->num_nuclei; i++) {

/* search for amino acid type */

    sscanf(data->atom_desc[i], "%s", amino_acid);
    ent.key = amino_acid;
    ep = hsearch(ent, FIND);
    if (ep == NULL) {
      if (params->verbosity >= 1) {
        fprintf(stderr, "Error: unknown amino acid '%s'.\n", amino_acid);
      }
      exit(1);
    }
    data->cs_res[i] = (int) ep->data;

/* enter full type */

    ent.key = data->atom_desc[i];
    ent.data = (char *) i;
    ep = hsearch(ent, FIND);
    if (ep == NULL) {
      ep = hsearch(ent, ENTER);
      if (ep == NULL) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error while building hash table for atom_desc.\n");
        }
        exit(1);
      }
    } else {
      if (params->verbosity >= 2) {
        fprintf(stdout, "Warning: type '%s' is duplicated.\n",
            data->atom_desc[i]);
      }
      data->num_nuclei--;
      for (j = i; j < data->num_nuclei; j++) {
        data->atom_desc[j] = data->atom_desc[j + 1];
      }
      i--;
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Setup hash table for amino acid types
 */
void init_htbl_AA3(Parameters *params)
{
  int i;
  ENTRY ent, *ep;

  for (i = 0; i < 20; i++) {
    ent.key = AA3[i];
    ent.data = (char *) i;
    ep = hsearch(ent, ENTER);
    if (ep == NULL) {
      if (params->verbosity >= 1) {
        fprintf(stderr, "Error while building hash table for AA3.\n");
      }
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void get_data(Data *data, Parameters *params)
{
  char buffer[MAX_LONG_BUF], file_name[MAX_BUF];
  char word[MAX_LONG_BUF], amino_acid[MAX_NAME_LEN], atom_name[MAX_NAME_LEN];
  char atom_type[MAX_NAME_LEN], atom_desc[16];
  FILE *fp;
  int atom_num, res_num, num_read, cs_num, last_res, res_auth_num;
  ENTRY ent, *ep;
  float cs_val;

  CALLOC2(data->cs_val, data->num_nuclei, MAX_CS_NUM, float);
  CALLOC2(data->cs_val_tag, data->num_nuclei, MAX_CS_NUM, int);
  CALLOC(data->cs_count, data->num_nuclei, int);
  CALLOC(data->res_count, 20, int);
  CALLOC(data->cs_mean, data->num_nuclei, float);

  while (fgets(buffer, MAX_LONG_BUF, stdin) != NULL) {
    if (buffer[0] == '#' || buffer[0] == '\n') {
      continue;
    }
    buffer[strlen(buffer) - 1] = '\0';
    sprintf(file_name, "%s/%s", params->BMRB_dir, buffer);
    if ((fp = fopen(file_name, "r")) == NULL) {
      fprintf(stderr, "Warning: unable to open file '%s'.\n", file_name);
      continue;
    }
    while (fgets(buffer, MAX_LONG_BUF, fp) != NULL) {
      sscanf(buffer, "%s", word);
      if (strncmp(word, "_Chem_shift_ambiguity_code", 26) == 0) {
        last_res = INT_MIN;
        while (strcmp(word, "stop_") != 0) {
          if (fgets(buffer, MAX_LONG_BUF, fp) != NULL) {
            num_read = sscanf(buffer, "%d%d%s%s%s%f", &atom_num, &res_num,
                amino_acid, atom_name, atom_type, &cs_val);
            if (num_read == 5) {
              num_read = sscanf(buffer, "%d%d%d%s%s%s%f", &atom_num,
                  &res_auth_num, &res_num, amino_acid, atom_name, atom_type,
                  &cs_val);
            }
            if (num_read == 6 || num_read == 7) {
              if (cs_val < MIN_CS_VAL || cs_val > MAX_CS_VAL) {
                fprintf(stderr, "Warning: strange chemical shift in '%s': "
                    "%d %d %s %s %s %f (skipping it)\n", file_name, atom_num,
                    res_num, amino_acid, atom_name, atom_type, cs_val);
              } else {
                if (params->group_atom_names == 0) {
                  sprintf(atom_desc, "%s %s %c", amino_acid, atom_name,
                    atom_type[0]);
                } else {
                  sprintf(atom_desc, "%s %.2s %c", amino_acid, atom_name,
                    atom_type[0]);
                }
                ent.key = atom_desc;
                ep = hsearch(ent, FIND);
                if (ep != NULL) {
                  cs_num = (int) ep->data;
                  data->cs_mean[cs_num] += cs_val;
                  data->cs_count[cs_num]++;
                  ent.key = amino_acid;
                  ep = hsearch(ent, FIND);
                  if (last_res != res_num) {
                    data->res_count[(int) ep->data]++;
                    last_res = res_num;
                  }
                  data->cs_val[cs_num]
                      [data->res_count[(int) ep->data] - 1] = cs_val;
                  data->cs_val_tag[cs_num]
                      [data->res_count[(int) ep->data] - 1] = 1;
                }
              }
            }
          } else {
            if (params->verbosity >= 1) {
              fprintf(stderr, "Error: unexpected End-Of-File\n");
            }
            exit(1);
          }
          sscanf(buffer, "%s", word);
        }
      }
    }
    FCLOSE(fp);
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void compute_stats(Data *data)
{
  int i, j;
  float min = 0, max = 0;

  ALLOC(data->cs_stdev, data->num_nuclei, float);
  ALLOC(data->cs_min, data->num_nuclei, float);
  ALLOC(data->cs_max, data->num_nuclei, float);
  ALLOC(data->cs_missing, data->num_nuclei, int);

/* find min, max and calculate mean */

  for (i = 0; i < data->num_nuclei; i++) {
    min = FLT_MAX;
    max = -FLT_MAX;
    for (j = 0; j < data->res_count[data->cs_res[i]]; j++) {
      if (data->cs_val_tag[i][j] != 0) {
        if (data->cs_val[i][j] < min) {
          min = data->cs_val[i][j];
        }
        if (data->cs_val[i][j] > max) {
          max = data->cs_val[i][j];
        }
      }
    }
    if (min < FLT_MAX - EPSILON) {
      data->cs_min[i] = min;
      data->cs_max[i] = max;
    } else {
      data->cs_min[i] = 0.0;
      data->cs_max[i] = 0.0;
    }
    if (data->cs_count[i] > 0) {
      data->cs_mean[i] /= data->cs_count[i];
    } else {
      data->cs_mean[i] = 0.0;
    }
  }

/* calculate standard deviations */

  for (i = 0; i < data->num_nuclei; i++) {
    data->cs_stdev[i] = 0.0;
    if (data->cs_count[i] != 0) {
      for (j = 0; j < data->res_count[data->cs_res[i]]; j++) {
        if (data->cs_val_tag[i][j] != 0) {
          data->cs_stdev[i] += (data->cs_val[i][j] - data->cs_mean[i]) *
              (data->cs_val[i][j] - data->cs_mean[i]);
        }
      }

/* (data->cs_count[i] - 1) is the non biased estimator */

      if (data->cs_count[i] > 1) {
        data->cs_stdev[i] /= data->cs_count[i] - 1;
      }
      data->cs_stdev[i] = sqrt(data->cs_stdev[i]);
    }
  }

/* calculate missing chemical shifts */

  for (i = 0; i < data->num_nuclei; i++) {
    data->cs_missing[i] = data->res_count[data->cs_res[i]] -
        data->cs_count[i];
  }

}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Print matrix in octave format to file pointer 'fp'
 */
void print_octave_matrix(FILE *fp, float **mat, int mat_dim)
{
  int i, j;

  fprintf(fp, "[");
  for (i = 0; i < mat_dim; i++) {
    fprintf(fp, "[");
    for (j = 0; j < mat_dim; j++) {
      fprintf(fp, "%12.4f", mat[i][j]);
    }
    fprintf(fp, "];");
  }
  fprintf(fp, "]\n\n");
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_stats(Data *data, Parameters *params)
{
  int i, div;
  char amino_acid[MAX_NAME_LEN], atom_name[MAX_NAME_LEN];
  char atom_type[MAX_NAME_LEN];
  float percentage_missing;
  FILE *fp;

  FOPEN(fp, params->out_fname, "w");

  fprintf(fp, "Amino   Atom    Atom     Number     Minimum     "
      "Maximum    Average    Standard    Missing   Percentage\n");
  fprintf(fp, "Acid    Name    Type    of Shifts    Shift       "
      "Shift      Shift     Deviation   Shifts     Missing\n\n");
  for (i = 0; i < data->num_nuclei; i++) {
    sscanf(data->atom_desc[i], "%s%s%s", amino_acid, atom_name, atom_type);
    if ((data->cs_count[i] + data->cs_missing[i]) == 0) {
      percentage_missing = 0.0;
    } else {
      div = data->cs_count[i] + data->cs_missing[i];
      if (div == 0) {
        div = 1;
      }
      percentage_missing = 100.0 * data->cs_missing[i] / div;
    }
    fprintf(fp, "%-8.8s%-9.9s%-9c%-6d%10.4f  %10.4f %10.4f "
        "%10.4f       %-6d%10.4f\n", amino_acid, atom_name, atom_type[0],
        data->cs_count[i], data->cs_min[i], data->cs_max[i], data->cs_mean[i],
        data->cs_stdev[i], data->cs_missing[i], percentage_missing);
  }

  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/

#define CUT_OFF 2.0

/**
 * Calculate covariances matrices
 * NOTE : ri and rj are only used for the cdata matrices.
 * NOTE2 : as covariations are not calculated on the same subsets, we have
 *         to set the diagonal elements to the larger value ever observed.
 */
void compute_covariances(Data *data, Parameters *params)
{
  int i, j, k, l, rj, rl, beg_row, end_row, cov_counts, size;
  float **cov, mean_j, mean_l, cov_j, cov_l;

  CALLOC(data->cdata, 20, Class_data);

  for (k = 0; k < 20; k++) {
    if (data->res_count[k] != 0) {
      beg_row = -1;
      end_row = -1;
      for (i = 0; i < data->num_nuclei; i++) {
        if (data->cs_res[i] == k) {
          if (beg_row == -1) {
            beg_row = i;
          }
          end_row = i;
        }
      }

#ifdef DEBUG
      printf("beg_row = %d -- end_row = %d\n", beg_row, end_row);
#endif

      size = data->cdata[k].size = end_row - beg_row + 1;
      CALLOC2(data->cdata[k].cov, size, size, float);
      if (params->print_cor_coefs == 1 || params->print_latex_cor_coefs == 1) {
        CALLOC2(data->cdata[k].cor_coef, size, size, float);
      }
      cov = data->cdata[k].cov;
      for (j = beg_row, rj = 0; j <= end_row; j++, rj++) {
        for (l = j + 1, rl = rj + 1; l <= end_row; l++, rl++) {
          mean_j = mean_l = cov_counts = 0.0;
          for (i = 0; i < data->res_count[k]; i++) {
            if (data->cs_val_tag[j][i] != 0 && data->cs_val_tag[l][i] != 0) {
              mean_j += data->cs_val[j][i];
              mean_l += data->cs_val[l][i];
              cov_counts++;
            }
          }
          if (cov_counts > 1) {
            mean_j /= cov_counts;
            mean_l /= cov_counts;

            cov_j = cov_l = 0.0;
            for (i = 0; i < data->res_count[k]; i++) {
              if (data->cs_val_tag[j][i] != 0 && data->cs_val_tag[l][i] != 0) {
                cov_j += (data->cs_val[j][i] - mean_j) *
                    (data->cs_val[j][i] - mean_j);
                cov_l += (data->cs_val[l][i] - mean_l) *
                    (data->cs_val[l][i] - mean_l);
              }
            }
            cov_j /= (cov_counts - 1);
            cov_l /= (cov_counts - 1);
            cov[rj][rj] = MAX2(cov[rj][rj], cov_j);
            cov[rl][rl] = MAX2(cov[rl][rl], cov_l);

            for (i = 0; i < data->res_count[k]; i++) {
              if (data->cs_val_tag[j][i] != 0 && data->cs_val_tag[l][i] != 0) {
                cov[rj][rl] += (data->cs_val[j][i] - mean_j) *
                    (data->cs_val[l][i] - mean_l);
              }
            }
            cov[rj][rl] /= (cov_counts - 1);
          } else {
            cov[rj][rl] = 0.0;
          }
          cov[rl][rj] = cov[rj][rl];
        } /* end loop on l = j to end_row       */
      }   /* end loop on j = beg_row to end_row */
    }     /* end if data->res_count[k] != 0     */
  }       /* end loop for k = 0 to 19           */
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_covariances(Data *data, Parameters *params)
{
  int i, j, k, size;
  float **cov;
  FILE *fp;

  FOPEN(fp, params->out_fname, "a");

  for (k = 0; k < 20; k++) {
    if (data->res_count[k] != 0) {
      fprintf(fp, "<cov res=\"%s\" num=\"%d\" dim=\"%d\">\n", AA3[k],
          data->res_count[k], data->cdata[k].size);
      cov = data->cdata[k].cov;
      size = data->cdata[k].size;
      for (i = 0; i < size; i++) {
        for (j = 0; j < size; j++) {
          fprintf(fp, "%18.10f", cov[i][j]);
        }
        fprintf(fp, "\n");
      }
      fprintf(fp, "</cov>\n");
    }
  }

  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Calculate correlations coefficients.
 * NOTE : ri and rj are only used for the cdata matrices.
 * IMPORTANTE NOTE: as covariations are not calculated on the same CS
 *                  subsets, WE HAVE TO recalculate cov_i and cov_j to
 *                  be consistent for correlation coefficients
 */
void calculate_correlation_coefficients(Data *data)
{
  int i, j, k, l, rj, rl, beg_row, end_row, cov_counts;
  float **cov, **cor_coef, div, cov_j, cov_l, mean_j, mean_l;

  for (k = 0; k < 20; k++) {
    if (data->res_count[k] != 0) {
      beg_row = -1;
      end_row = -1;
      for (i = 0; i < data->num_nuclei; i++) {
        if (data->cs_res[i] == k) {
          if (beg_row == -1) {
            beg_row = i;
          }
          end_row = i;
        }
      }

      cov = data->cdata[k].cov;
      cor_coef = data->cdata[k].cor_coef;
      for (j = beg_row, rj = 0; j <= end_row; j++, rj++) {
        for (l = j, rl = rj; l <= end_row; l++, rl++) {
          mean_j = mean_l = cov_counts = 0.0;
          for (i = 0; i < data->res_count[k]; i++) {
            if (data->cs_val_tag[j][i] != 0 && data->cs_val_tag[l][i] != 0) {
              mean_j += data->cs_val[j][i];
              mean_l += data->cs_val[l][i];
              cov_counts++;
            }
          }
          if (cov_counts > 1) {
            mean_j /= cov_counts;
            mean_l /= cov_counts;

            cov_j = cov_l = 0.0;
            for (i = 0; i < data->res_count[k]; i++) {
              if (data->cs_val_tag[j][i] != 0 && data->cs_val_tag[l][i] != 0) {
                cov_j += (data->cs_val[j][i] - mean_j) *
                    (data->cs_val[j][i] - mean_j);
                cov_l += (data->cs_val[l][i] - mean_l) *
                    (data->cs_val[l][i] - mean_l);
              }
            }
            cov_j /= (cov_counts - 1);
            cov_l /= (cov_counts - 1);

            div = sqrt(cov_j * cov_l);
            if (div != 0) {
              cor_coef[rj][rl] = cov[rj][rl] / div;
            } else {
              cor_coef[rj][rl] = 0.0;
            }
          } else {
            cor_coef[rj][rl] = 0.0;
          }
          cor_coef[rl][rj] = cor_coef[rj][rl];
        } /* end loop on l = j to end_row       */
      }   /* end loop on j = beg_row to end_row */
    }     /* end if data->res_count[k] != 0     */
  }       /* end loop for k = 0 to 19           */
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Print correlations coefficients in a "pseudo" XML format
 */
void print_correlation_coefficients(Data *data, Parameters *params)
{
  int i, j, k, size, beg_row;
  float **cor_coef;
  char atom_type[16];
  FILE *fp;

  FOPEN(fp, params->out_fname, "a");

  for (k = 0; k < 20; k++) {
    if (data->res_count[k] != 0) {
      beg_row = -1;
      for (i = 0; i < data->num_nuclei; i++) {
        if (data->cs_res[i] == k) {
          if (beg_row == -1) {
            beg_row = i;
          }
        }
      }
      cor_coef = data->cdata[k].cor_coef;
      size = data->cdata[k].size;
      fprintf(fp, "<cor_coef res=\"%s\" dim=\"%d\">\n", AA3[k], size);
      fprintf(fp, "          ");
      for (i = 0; i < size; i++) {
        sscanf(data->atom_desc[i + beg_row], "%s%s", atom_type, atom_type);
        fprintf(fp, "%10.10s", atom_type);
      }
      fprintf(fp, "\n");
      for (i = 0; i < size; i++) {
        sscanf(data->atom_desc[i + beg_row], "%s%s", atom_type, atom_type);
        fprintf(fp, "%10.10s", atom_type);
        for (j = 0; j < size; j++) {
          fprintf(fp, "%10.4f", cor_coef[i][j]);
        }
        fprintf(fp, "\n");
      }
      fprintf(fp, "</cor_coef>\n");
      /*print_octave_matrix(stdout, cor_coef, size);*/
    }
  }

  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/

#define XSCALE 0.35
#define YSCALE 0.35

void print_latex_cor_coefs(Data *data)
{
  int i, j, k, ri, rj, size, shift;
  float **cor_coef;
  char atom_type[16];
  FILE *fp;

  FOPEN(fp, "BMRB_stats.tex", "w");

  for (k = 0; k < 20; k++) {
    shift = 0;
    if (data->res_count[k] != 0) {
      size = data->cdata[k].size;
      fprintf(fp, "\\begin{pspicture}(%.2f,%.2f)(%.2f,%.2f)\n",
          -2.0 * XSCALE, -2.0 * YSCALE, XSCALE * size, YSCALE * (size + 2));
      fprintf(fp, "\\psset{xunit=%.2f}\n", XSCALE);
      fprintf(fp, "\\psset{yunit=%.2f}\n", YSCALE);
      fprintf(fp, "\\psset{fillstyle=solid,linecolor=white}\n");
      cor_coef = data->cdata[k].cor_coef;
      for (i = 0, ri = 0; i < data->num_nuclei; i++) {
        if (data->cs_res[i] == k) {
          if (ri > 0) {
            sscanf(data->atom_desc[i], "%s%s", atom_type, atom_type);
            fprintf(fp, "\\rput(%f,%f){\\tiny %s}\n", -0.5 + ri,
                0.5 + size + shift, atom_type);
            sscanf(data->atom_desc[i - 1], "%s%s", atom_type, atom_type);
            fprintf(fp, "\\rput[r](%f,%f){\\tiny %s}\n", -0.5,
                0.5 + size - ri, atom_type);
            if (shift != 0) {
              shift = 0;
            } else {
              shift++;
            }
          }
          for (j = i + 1, rj = ri + 1; j < data->num_nuclei; j++) {
            if (data->cs_res[j] == k) {
              if (cor_coef[ri][rj] >= 0) {
                fprintf(fp, "\\newrgbcolor{AAAA}{%.2f %.2f 1}\n",
                    1.0 - cor_coef[ri][rj], 1.0 - cor_coef[ri][rj]);
              } else {
                fprintf(fp, "\\newrgbcolor{AAAA}{1 %.2f %.2f}\n",
                    1.0 + cor_coef[ri][rj], 1.0 + cor_coef[ri][rj]);
              }
              fprintf(fp, "\\psframe[fillcolor=AAAA](%d,%d)(%d,%d)\n",
                  rj - 1, size - ri - 1, rj, size - ri);
              rj++;
            }
          }
          ri++;
        }
      }
      fprintf(fp, "\\rput(%.2f,-1){%s}\n", (float) size / 2.0, AA3[k]);
      fprintf(fp, "\\end{pspicture}\n");
    }
  }

  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int main(int argc, char **argv)
{
  Parameters params;
  Data data;

  init_parameters(&params);
  process_command_line(argc, argv, &params);
  check_parameters(&params, argv);

  get_atoms_to_analyse(&params, &data);

  hcreate(4 * data.num_nuclei + 40);
  init_htbl_AA3(&params);
  init_htbl(&data, &params);

  get_data(&data, &params);

  compute_stats(&data);
  print_stats(&data, &params);

  compute_covariances(&data, &params);
  print_covariances(&data, &params);

  if (params.print_cor_coefs == 1 || params.print_latex_cor_coefs == 1) {
    calculate_correlation_coefficients(&data);
  }
  if (params.print_cor_coefs == 1) {
    print_correlation_coefficients(&data, &params);
  }
  if (params.print_latex_cor_coefs == 1) {
    print_latex_cor_coefs(&data);
  }

  return 0;
}
