/*
 * This progam extracts chemical shifts from BMRB files.
 *
 * Copyright (C) 2003 Antoine MARIN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1 published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact the author via mail or email at:
 * Antoine Marin (Antoine.Marin@ibpc.fr)
 * CNRS UPR 9080
 * Laboratoire de Biochimie Thorique
 * 13, rue Pierre et Marie Curie
 * 75005 Paris - France
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <search.h> /* for hash tables */
#include <libgen.h> /* for basename */
#include <ctype.h>  /* for toupper */

#include "alloc_utils.h"

#define VERSION "1.0"

/* Default values */

#define TYPES_FILE "cs.in"
#define STAT_FILE "BMRB_stats.out"
#define DEF_MUL_FAC 3.0
#define DEF_VERB 2

/* MIN and MAX values */

#define MAX_BUF 1024
#define MAX_ATOM_TYPES 150
#define MAX_RES_TYPES 20
#define MAX_NAME_LEN 16

#define MIN_MUL_FAC 0.0
#define MAX_MUL_FAC 10.0
#define MIN_VERB 0
#define MAX_VERB 3
#define MIN_PH 0.0
#define MAX_PH 14.0
#define MIN_TEMP 273.0
#define MAX_TEMP 373.0

char *AA3[] = {"ALA", "ARG", "ASP", "ASN", "CYS", "GLU", "GLN", "GLY", "HIS",
 "ILE", "LEU", "LYS", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL"
};

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
typedef struct parameters {
  int filter;              /**< if = 1, filter chemical shifts with pH,
                                temerature and mean values */
  int verbosity;           /**< verbosity level */
  int only_cs_num;         /**< if = 1, print only the number of chemical
                                shifts read */
  int min_cs_to_print;     /**< minimum number of chemical shifts to print the
                                corresponding residue
                                NOTE: there are 3 different cases:
                                1) if the value is > 0 then at least this
                                   number of chemical shifts is required to
                                   print the residue
                                2) if the value is equal to 0 then the
                                   exact number of different chemical shifts
                                   read in the TYPES_FILE file is required
                                3) if the value is < 0 then at least the number
                                   of different chemical shifts read in the
                                   TYPES_FILE file plus this number is
                                   required */
  float factor;            /**< factor multiplying standard deviation
                                to filter chemical shifts (if required) */
  float T_min, T_max;      /**< min and max temperatures to filter chemical
                                shifts */
  float pH_min, pH_max;    /**< min and max pH to filter chemical shifts */
  char *cs_input_fname;    /**< chemical shifts input types */
  char *stat_input_fname;  /**< BMRB statistics file to filter chemical
                                shifts */
} Parameters;

typedef struct data {
  int *max_atoms;   /**< maximum number of atom names per residue */
  int *stat_flag;   /**< flag to check what statistic was read */
  int num_keys;     /**< number of different keys */
  char **keys;      /**< atom keys for the hash table = amino_acid +
                         atom_name + atom_type */
  float *means;     /**< means for each key */
  float *std_devs;  /**< standard deviations for each key */
} Data;

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void init_parameters(Parameters *params)
{
  params->filter = 0;            /* do NOT filter by default */
  params->verbosity = DEF_VERB;
  params->only_cs_num = 0;
  params->min_cs_to_print = 1;   /* by default print chemical shifts when
                                    there is at least 1 chemical shift
                                    available */
  params->factor = DEF_MUL_FAC;
  params->T_min = MIN_TEMP;
  params->T_max = MAX_TEMP;
  params->pH_min = MIN_PH;
  params->pH_max = MAX_PH;
  params->cs_input_fname = TYPES_FILE;
  params->stat_input_fname = STAT_FILE;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_usage(FILE *fp, char **argv, Parameters *params)
{
  fprintf(fp, "\nUsage: %s [options] -ics file < BMRB_file.str\n\n",
      basename(argv[0]));
  fprintf(fp, "  Options:\n\n");
  fprintf(fp, "    -f            : filter chemical shifts according to "
      "standard values,\n");
  fprintf(fp, "                    temperature and pH.\n\n");
  fprintf(fp, "    -h,--help     : print this help.\n\n");
  fprintf(fp, "    -ics file     : file containing chemical shifts to "
      "take into account (mandatory, default %s).\n\n", TYPES_FILE);
  fprintf(fp, "    -ist file     : file containing BMRB statistics "
      "to filter chemical shifts (default %s).\n\n", STAT_FILE);
  fprintf(fp, "    -mc int       : minimum number of chemical shifts to "
      "print (default = %d).\n", params->min_cs_to_print);
  fprintf(fp, "                    NOTE: there are 3 different cases:\n");
  fprintf(fp, "                 1) if the value is > 0 then at least this\n");
  fprintf(fp, "                    number of chemical shifts is required to\n");
  fprintf(fp, "                    print the residue.\n");
  fprintf(fp, "                 2) if the value is equal to 0 then the\n");
  fprintf(fp, "                    exact number of different chemical shifts\n");
  fprintf(fp, "                    read in the file '%s' is required.\n",
      TYPES_FILE);
  fprintf(fp, "                 3) if the value is < 0 then at least the number\n");
  fprintf(fp, "                    of different chemical shifts read in the\n");
  fprintf(fp, "                    file '%s' plus this number is required.\n\n",
      TYPES_FILE);
  fprintf(fp, "    -mf float     : float is the factor multiplying "
      "the standard deviation\n"
      "                    used to cut-off chemical shifts "
      "(default = %.1f), range: [%.1f - %.1f].\n\n",
      DEF_MUL_FAC, MIN_MUL_FAC, MAX_MUL_FAC);
  fprintf(fp, "    -n            : only count the number of chemical "
      "shifts in the file.\n\n");
  fprintf(fp, "    -pH_max float : maximum pH below which chemical shifts are "
      "printed\n                    (default = %.1f), range: "
      "[%.1f - %.1f].\n\n", MAX_PH, MIN_PH, MAX_PH);
  fprintf(fp, "    -pH_min float : minimum pH above which chemical shifts are "
      "printed\n                    (default = %.1f), range: "
      "[%.1f - %.1f].\n\n", MIN_PH, MIN_PH, MAX_PH);
  fprintf(fp, "    -T_max float  : maximum temperature (in Kelvin) "
      "below which chemical shifts are printed\n"
      "                    (default = %.1f), range: [%.1f - %.1f].\n\n",
      MAX_TEMP, MIN_TEMP, MAX_TEMP);
  fprintf(fp, "    -T_min float  : minimum temperature (in Kelvin) "
      "above which chemical shifts are printed\n"
      "                    (default = %.1f), range: [%.1f - %.1f].\n\n",
      MIN_TEMP, MIN_TEMP, MAX_TEMP);
  fprintf(fp, "    -v, --version : print version only (current version = %s).\n\n",
      VERSION);
  fprintf(fp, "    -verb int     : verbosity level (default = %d).\n",
      DEF_VERB);
  fprintf(fp, "                    0 -> quiet\n");
  fprintf(fp, "                    1 -> errors\n");
  fprintf(fp, "                    2 -> errors + warnings\n");
  fprintf(fp, "                    3 -> errors + warnings + more "
      "informations\n\n");
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void process_command_line(int argc, char **argv, Parameters *params)
{
  int i;

  for (i = 1; i < argc; i++) {

    if (strcmp(argv[i], "-f") == 0) {

/* -f : filter cs values according to standard deviations */

      params->filter = 1;
    } else if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {

/* -h, --help : print the usage message */

      print_usage(stdout, argv, params);
      exit(0);
    } else if (strcmp(argv[i], "-ics") == 0) {

/* -ics : input chemical shifts file */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-ics' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      ALLOC(params->cs_input_fname, strlen(argv[i]) + 1, char);
      strcpy(params->cs_input_fname, argv[i]);
    } else if (strcmp(argv[i], "-ist") == 0) {

/* -ist : input statistics file */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-ist' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      ALLOC(params->stat_input_fname, strlen(argv[i]) + 1, char);
      strcpy(params->stat_input_fname, argv[i]);
    } else if (strcmp(argv[i], "-mc") == 0) {

/* -mc : minimum number of chemical shifts to print */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-mc' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      params->min_cs_to_print = atoi(argv[i]);
    } else if (strcmp(argv[i], "-mf") == 0) {

/* -mf : multiplying factor */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-mf' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      params->factor = atof(argv[i]);
      if (params->factor < MIN_MUL_FAC || params->factor > MAX_MUL_FAC) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: factor value out of range: %.1f\n",
              params->factor);
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
    } else if (strcmp(argv[i], "-n") == 0) {

/* -n : only count the number of chemical shifts */

      params->only_cs_num = 1;
    } else if (strcmp(argv[i], "-pH_max") == 0) {

/* -pH_max : maximum pH to print chemical shifts */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-pH_max' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      params->pH_max = atof(argv[i]);
      if (params->pH_max < MIN_PH || params->pH_max > MAX_PH) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: pH_max value out of range: %.1f\n",
              params->pH_max);
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
    } else if (strcmp(argv[i], "-pH_min") == 0) {

/* -pH_min : minimum pH to print chemical shifts */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-pH_min' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      params->pH_min = atof(argv[i]);
      if (params->pH_min < MIN_PH || params->pH_min > MAX_PH) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: pH_min value out of range: %.1f\n",
              params->pH_min);
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
    } else if (strcmp(argv[i], "-T_max") == 0) {

/* -T-max : maximum temperature to print chemical shifts */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-T_max' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      params->T_max = atof(argv[i]);
      if (params->T_max < MIN_TEMP || params->T_max > MAX_TEMP) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: T_max value out of range: %.1f\n",
              params->T_max);
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
    } else if (strcmp(argv[i], "-T_min") == 0) {

/* -T_min : minimum temperature to print chemical shifts */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-T_min' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      params->T_min = atof(argv[i]);
      if (params->T_min < MIN_TEMP || params->T_min > MAX_TEMP) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: T_min value out of range: %.1f\n",
              params->T_min);
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
    } else if (strcmp(argv[i], "-v") == 0 ||
      strcmp(argv[i], "--version") == 0) {

/* -v, --version : print version and exit */

      fprintf(stdout, "This is %s release number %s\n", basename(argv[0]),
          VERSION);
      exit(0);
    } else if (strcmp(argv[i], "-verb") == 0) {

/* -verb : verbosity */

      i++;
      if (i >= argc) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error: you should provide a value for the "
              "'-verb' option.\n");
          print_usage(stderr, argv, params);
        }
        exit(1);
      }
      params->verbosity = atoi(argv[i]);
      if (params->verbosity < MIN_VERB || params->verbosity > MAX_VERB) {
        fprintf(stderr, "Error: verbosity level out of range: %d\n",
            params->verbosity);
        print_usage(stderr, argv, params);
        exit(1);
      }
    } else {

/* Unknown tag */

      if (params->verbosity >= 1) {
        fprintf(stderr, "Error: unknown tag: '%s'\n", argv[i]);
        print_usage(stderr, argv, params);
      }
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Setup hash table for amino acid types
 */
void init_htbl_AA3(Parameters *params)
{
  int i;
  ENTRY ent, *ep;

  for (i = 0; i < MAX_RES_TYPES; i++) {
    ent.key = AA3[i];
    ent.data = (char *) i;
    ep = hsearch(ent, ENTER);
    if (ep == NULL) {
      if (params->verbosity >= 1) {
        fprintf(stderr, "Error while building hash table for AA3.\n");
      }
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Setup hash table for atom types
 */
void read_input_types(Data *data, Parameters *params)
{
  ENTRY ent, *ep;
  int nel;
  char buffer[MAX_BUF], amino_acid[MAX_NAME_LEN], atom_name[MAX_NAME_LEN],
      atom_type[MAX_NAME_LEN];
  FILE *fp;

  FOPEN(fp, params->cs_input_fname, "r");

  CALLOC(data->max_atoms, MAX_RES_TYPES, int);

/* get number of keys and count number of atom names per residu */

  nel = 0;
  while (fgets(buffer, MAX_BUF, fp) != NULL) {
    if (buffer[0] != '#' && buffer[0] != '\n') {
      sscanf(buffer, "%s", amino_acid);
      ent.key = amino_acid;
      ep = hsearch(ent, FIND);
      if (ep == NULL) {
        fprintf(stderr, "Error: unknown residue '%s' line %d in file '%s'\n",
            amino_acid, nel + 1, params->cs_input_fname);
        exit(1);
      }
      data->max_atoms[(int) ep->data]++;
      nel++;
    }
  }
  ALLOC2(data->keys, nel, 3 * MAX_NAME_LEN, char);
  ALLOC(data->means, nel, float);
  ALLOC(data->std_devs, nel, float);
  CALLOC(data->stat_flag, nel, int);
  data->num_keys = nel;

/* get keys */

  rewind(fp);
  nel = 0;
  while (fgets(buffer, MAX_BUF, fp) != NULL) {
    if (buffer[0] != '#' && buffer[0] != '\n') {
      sscanf(buffer, "%s%s%s", amino_acid, atom_name, atom_type);
      sprintf(data->keys[nel], "%s %s %s", amino_acid, atom_name, atom_type);
      ent.key = data->keys[nel];
      ent.data = (char *) nel;
      ep = hsearch(ent, ENTER);
      if (ep == NULL) {
        if (params->verbosity >= 1) {
          fprintf(stderr, "Error while building hash table for atom types.\n");
        }
        exit(1);
      }
      nel++;
    }
  }

  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
float C_to_K(float temp)
{
  return 273.0 + temp;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
float F_to_K(float temp)
{
  return 273.0 + ((temp - 32.0) * 5.0) / 9.0;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int check_ph_temp(Parameters *params)
{
  char buffer[MAX_BUF], word[MAX_BUF], unit[32], temp_error[32];
  int num_read;
  float temp, pH;

  while (fgets(buffer, MAX_BUF, stdin) != NULL) {

/* AAARRGGGHHH there is at least 25 different names for this field ! :
 * - save_sample_condition_set_one
 * - save_condition_1
 * - save_sample_conditions_one
 * - save_Cond_1
 * - save_sample_conditions
 * - save_Experimental_conditions_1
 * - save_ss1559-cond_1
 * - ...
 *
 * I grouped together the 'save_sample_condition' ones,
 * the 'save_condition' ones and the 'save_Cond' ones.
 * I finally added the preceeding comment that is much better conserved
 * than any other field name !!!!!?????
 */

    if (strncmp(buffer, "#  Sample conditions  #", 23) == 0 ||
        strncmp(buffer, "save_sample_condition", 21) == 0 ||
        strncmp(buffer, "save_condition", 14) == 0 ||
        strncmp(buffer, "save_Cond", 9) == 0) {
      fgets(buffer, MAX_BUF, stdin);
      sscanf(buffer, "%s", word);
      while (strcmp(word, "loop_") != 0) {
        if (fgets(buffer, MAX_BUF, stdin) == NULL) {
          if (params->verbosity >= 1) {
            fprintf(stderr, "Error: unexpected End-Of-File\n");
          }
          exit(1);
        }
        sscanf(buffer, "%s", word);
      }
      fgets(buffer, MAX_BUF, stdin);
      sscanf(buffer, "%s", word);
      while (strcmp(word, "stop_") != 0) {
        if (fgets(buffer, MAX_BUF, stdin) != NULL) {
          sscanf(buffer, "%s", word);
          if (strcmp(word, "pH") == 0 || strcmp(word, "pH*") == 0) {

/* Get pH */

            if (sscanf(buffer, "%s%f", word, &pH) != 2) {
              if (params->verbosity >= 1) {
                fprintf(stderr, "Warning: unable to find pH, skipping sequence.\n");
              }
              if (params->verbosity >= 3) {
                fprintf(stderr, "buffer:\n%s", buffer);
              }
              return 1;
            }
          } else if (strcmp(word, "temperature") == 0) {

/* Get temperature */

            num_read = sscanf(buffer, "%s%f%s%s", word, &temp, temp_error,
                unit);
            unit[0] = toupper(unit[0]);
            if (num_read == 4) {
              if (unit[0] == 'C') {
                temp = C_to_K(temp);
              } else if (unit[0] == 'F') {
                temp = F_to_K(temp);
              } else if (unit[0] != 'K') {
                if (params->verbosity >= 1) {
                  fprintf(stderr, "Warning: unknown temperature unit: %s, "
                      "skipping sequence.\n", unit);
                }
                if (params->verbosity >= 3) {
                  fprintf(stderr, "buffer:\n%s", buffer);
                }
                return 1;
              }
            } else if (num_read == 3) {

/*
 * this is a trick to overcome misformed files where the error is missing
 * between the temperature and the unit.
 * (ex: 'temperature 298 K' instead of '298 . K' or '298 0.2 K')
 */

              temp_error[0] = toupper(temp_error[0]);
              if (temp_error[0] == 'C') {
                temp = C_to_K(temp);
              } else if (temp_error[0] == 'F') {
                temp = F_to_K(temp);
              } else if (temp_error[0] != 'K') {
                if (params->verbosity >= 1) {
                  fprintf(stderr, "Warning: unknown temperature unit: %s, "
                      "skipping sequence.\n", unit);
                }
                if (params->verbosity >= 3) {
                  fprintf(stderr, "buffer:\n%s", buffer);
                }
                return 1;
              }
            } else {
              if (params->verbosity >= 1) {
                fprintf(stderr, "Warning: unable to find temperature, "
                    "skipping sequence.\n");
              }
              if (params->verbosity >= 3) {
                fprintf(stderr, "buffer:\n%s", buffer);
              }
              return 1;
            }
            break;
          }
        } else {
          if (params->verbosity >= 1) {
            fprintf(stderr, "Error: unexpected End-Of-File\n");
          }
          exit(1);
        }
      }
    }
  }
  if (params->verbosity >= 3) {
    fprintf(stderr, "T = %.1f [%.1f - %.1f], pH = %.1f [%.1f - %.1f]\n",
        temp, params->T_min, params->T_max, pH, params->pH_min,
        params->pH_max);
  }
  if (temp < params->T_min || temp > params->T_max ||
      pH < params->pH_min || pH > params->pH_max) {
    return 1;
  }
  return 0;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Print chemical shifts for one residue in XML format
 */
void print_cs(char *res, int nH, float *Hcs, int nN, float *Ncs, int nC,
    float *Ccs, Parameters *params, Data *data, int res_num)
{
  int i;
/*
  printf("res = %s, nH = %d, nC = %d, nN = %d, res_num = %d, max_atoms = %d\n",
      res, nH, nC, nN, res_num, data->max_atoms[res_num]);
*/
/* min_cs_to_print > 0 */

  if ((params->min_cs_to_print > 0 &&
      (nH + nC + nN) >= params->min_cs_to_print) ||
      (params->min_cs_to_print == 0 &&
      (nH + nC + nN) == data->max_atoms[res_num]) ||
      (params->min_cs_to_print < 0 &&
      (nH + nC + nN) >= data->max_atoms[res_num] + params->min_cs_to_print)) {
      printf("<res type=\"%s\">\n", res);

/* Hydrogen */

    if (nH != 0) {
      printf("  <cs type=\"H\" n=\"%d\">\n    ", nH);
      for (i = 0; i < nH; i++) {
        printf("%g ", Hcs[i]);
      }
      printf("\n  </cs>\n");
    }

/* Azote */

    if (nN != 0) {
      printf("  <cs type=\"N\" n=\"%d\">\n    ", nN);
      for (i = 0; i < nN; i++) {
        printf("%g ", Ncs[i]);
      }
      printf("\n  </cs>\n");
    }

/* Carbon */

    if (nC != 0) {
      printf("  <cs type=\"C\" n=\"%d\">\n    ", nC);
      for (i = 0; i < nC; i++) {
        printf("%g ", Ccs[i]);
      }
      printf("\n  </cs>\n");
    }
    printf("</res>\n");
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void extract_cs(Data *data, Parameters *params)
{
  char buffer[MAX_BUF], word[MAX_BUF], amino_acid[MAX_NAME_LEN];
  char atom_name[MAX_NAME_LEN], atom_type[MAX_NAME_LEN], key[MAX_NAME_LEN];
  char last_amino_acid[MAX_NAME_LEN];
  ENTRY ent, *ep;
  float Hcs[MAX_NAME_LEN], Ncs[MAX_NAME_LEN], Ccs[MAX_NAME_LEN], cs_val;
  int res_num, res_class, last_res_num, atom_num, num_read, flag;
  int nH, nN, nC, total, res_auth_num, key_num;

  last_res_num = 0;
  nH = nN = nC = 0;
  flag = total = 0;
  key_num = 0;
  res_class = -1;
  while (fgets(buffer, MAX_BUF, stdin) != NULL) {
    sscanf(buffer, "%s", word);
    if (strncmp(word, "_Chem_shift_ambiguity_code", 26) == 0) {
      while (strcmp(word, "stop_") != 0) {
        if (fgets(buffer, MAX_BUF, stdin) != NULL) {
          num_read = sscanf(buffer, "%d%d%s%s%s%f", &atom_num, &res_num,
              amino_acid, atom_name, atom_type, &cs_val);
          if (num_read == 5) {
            num_read = sscanf(buffer, "%d%d%d%s%s%s%f", &atom_num,
                &res_auth_num, &res_num, amino_acid, atom_name, atom_type,
                &cs_val);
          }
          if (num_read == 6 || num_read == 7) {
            sprintf(key, "%s %s %s", amino_acid, atom_name, atom_type);
            ent.key = key;
            ep = hsearch(ent, FIND);
            if (ep != NULL) { /* key has been found */
              key_num = (int) ep->data;
              if (last_res_num != res_num) { /* it is a new residue */
                if (params->only_cs_num == 0) {
                  if (last_res_num != 0 && flag == 1) {
                    print_cs(last_amino_acid, nH, Hcs, nN, Ncs, nC, Ccs, params,
                        data, res_class);
                  }
                } else {
                  total += nH + nN + nC;
                }
                nH = nN = nC = 0;
                ent.key = amino_acid;
                ep = hsearch(ent, FIND);
                if (ep == NULL) { /* residue is unknown */
                  if (params->verbosity >= 2) {
                    fprintf(stderr, "Warning: unknown residue type: %s\n",
                        amino_acid);
                  }
                  flag = 0;
                } else {
                  strcpy(last_amino_acid, amino_acid);
                  res_class = (int) ep->data;
                  last_res_num = res_num;
                  flag = 1;
                }
              }
              if (flag == 1) {
                if (params->filter == 1) {
                  if ((cs_val < (data->means[key_num] -
                    (params->factor * data->std_devs[key_num])) ||
                    cs_val > (data->means[key_num] +
                    (params->factor * data->std_devs[key_num])))) {
                    ep = NULL;
                  }
                }
                if (params->filter == 0 || ep != NULL) {
                  if (atom_type[0] == 'H') {
                    Hcs[nH] = cs_val;
                    nH++;
                  } else if (atom_type[0] == 'N') {
                    Ncs[nN] = cs_val;
                    nN++;
                  } else if (atom_type[0] == 'C') {
                    Ccs[nC] = cs_val;
                    nC++;
                  } else {
                    if (params->verbosity >= 1) {
                      fprintf(stderr, "Error: unknown atom type '%c'\n",
                          atom_type[0]);
                    }
                    exit(1);
                  }
                } else if (params->verbosity >= 3) {
                  printf("atom number = %d, cs = %f\n", atom_num, cs_val);
                }
              }
            }
          }
        } else {
          if (params->verbosity >= 1) {
            fprintf(stderr, "Error: unexpected End-Of-File\n");
          }
          exit(1);
        }
        sscanf(buffer, "%s", word);
      }
    }
  }
  if (params->only_cs_num == 0) {
    if (flag == 1) {
      print_cs(last_amino_acid, nH, Hcs, nN, Ncs, nC, Ccs, params, data,
          res_class);
    }
  } else {
    if (flag == 1) {
      total += nH + nN + nC;
    }
    printf("%d\n", total);
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Read statistics on chemical shifts in order to be able to filter them later
 */
void read_standard_values(Data *data, Parameters *params)
{
  char amino_acid[MAX_NAME_LEN], atom_name[MAX_NAME_LEN],
      atom_type[MAX_NAME_LEN], buffer[MAX_BUF], key[3 * MAX_NAME_LEN];
  int i, num_read, num_shifts;
  float min_shift, max_shift, mean, std_dev;
  FILE *fp;
  ENTRY ent, *ep;

  FOPEN(fp, params->stat_input_fname, "r");

  fgets(buffer, MAX_BUF, fp);
  fgets(buffer, MAX_BUF, fp);
  while (fgets(buffer, MAX_BUF, fp) != NULL) {
    if (strncmp(buffer, "<cov ", 5) == 0) {
      break;
    }
    num_read = sscanf(buffer, "%s%s%s%d%f%f%f%f", amino_acid, atom_name,
        atom_type, &num_shifts, &min_shift, &max_shift, &mean, &std_dev);
    if (num_read == 8) {
      sprintf(key, "%s %s %s", amino_acid, atom_name, atom_type);
      ent.key = key;
      ep = hsearch(ent, FIND);
      if (ep != NULL) {
        data->means[(int) ep->data] = mean;
        data->std_devs[(int) ep->data] = std_dev;
        data->stat_flag[(int) ep->data]++;
      }
    }
  }

/* check that all and not too much statistics have been read */

  for (i = 0; i < data->num_keys; i++) {
    if (data->stat_flag[i] == 0) {
      if (params->verbosity >= 1) {
        fprintf(stderr, "Error: no statistics were read for key number %d "
            "(%s) in file '%s'.\n", i + 1, data->keys[i],
            params->stat_input_fname);
      }
      exit(1);
    } else if (data->stat_flag[i] > 1) {
      if (params->verbosity >= 1) {
        fprintf(stderr, "Error: multiple statistics were read (%d) for key "
            "number %d (%s) in file '%s'.\n", data->stat_flag[i], i + 1,
            data->keys[i], params->stat_input_fname);
      }
      exit(1);
    }
  }

  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int main(int argc, char **argv)
{
  Parameters params;
  Data data;

  init_parameters(&params);
  process_command_line(argc, argv, &params);

/* check pH and temperature */

  if (params.filter == 1) {
    if (check_ph_temp(&params) != 0) {
      if (params.only_cs_num == 1) {
        printf("0\n");
      }
      exit(1);
    }
  }

/* create hash table */

  hcreate(4 * MAX_RES_TYPES * MAX_ATOM_TYPES);
  init_htbl_AA3(&params);

/* read atom types to be selected */

  read_input_types(&data, &params);

  if (params.filter == 1) {
    read_standard_values(&data, &params);
  }

/* extract chemical shifts */

  rewind(stdin);
  extract_cs(&data, &params);

  return 0;
}
