#!/bin/bash

# This script permits to group BMRB sequences into homologous groups using
# the FASTA program.
#
# Copyright (C) 2003 Antoine MARIN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License, version 2.1 published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# You can contact the author via mail or email at:
# Antoine Marin (Antoine.Marin@ibpc.fr)
# CNRS UPR 9080
# Laboratoire de Biochimie Thorique
# 13, rue Pierre et Marie Curie
# 75005 Paris - France

################################################################################
#
################################################################################
#
# Default values
#
VERSION="1.0"
MIN_PROT_LEN=10
# Directories
WORK_DIR=`pwd`
BMRB_DIR="/home/miro/marin/BMRB"
FASTA_SEQ_DIR="${BMRB_DIR}/fasta"
# Files
LOG_FILE="BMRB_to_homol_groups.log"
BMRB_PROT_FILES="BMRB_proteins_files.txt"
BMRB_NOMAG_PROT_FILES="BMRB_nomag_proteins_files.txt"
FASTA_PROT_FILES="BMRB_proteins_fasta_files.txt"
FASTA_FILTERED_FILES="BMRB_proteins_fasta_filtered_files.txt"
DIST_MAT="BMRB_proteins_distance_matrix.txt"
PROT_NAME_FILE="BMRB_proteins_names.txt"
GROUPS_FILE="BMRB_proteins_groups.txt"
COUNT_CS_FILE="BMRB_proteins_groups_cs.txt"
FINAL_PROT_FILE="BMRB_selected_sequences.txt"

################################################################################
#
################################################################################
print_help()
{
  echo ""
  echo "Usage: ${0##*/} [options]"
  echo ""
  echo "  Options:"
  echo ""
  echo "    --clean       : remove generated files"
  echo ""
  echo "    -d directory  : directory containing BMRB files to cluster"
  echo "                    (default = $BMRB_DIR)"
  echo ""
  echo "    -fd directory : output directory for fasta files"
  echo "                    (default = $FASTA_SEQ_DIR)"
  echo ""
  echo "    -h, --help    : print this help"
  echo ""
  echo "    -m int        : minimum sequence length to be selected"
  echo "                    (default = $MIN_PROT_LEN)"
  echo ""
  echo "    -v, --version : print version only (current version = $VERSION)"
  echo ""
}

################################################################################
#
################################################################################
clean()
{
  \rm -f $BMRB_PROT_FILES $FASTA_PROT_FILES $FASTA_FILTERED_FILES
  \rm -f $LOG_FILE $DIST_MAT $PROT_NAME_FILE $GROUPS_FILE $COUNT_CS_FILE
  \rm -f $BMRB_NOMAG_PROT_FILES
  align_all_vs_all.sh -fl $FASTA_FILTERED_FILES --clean
}

################################################################################
#
################################################################################
#
# Process command line
#
while test -n "`echo $1 | grep '-'`"; do
  case $1 in
    "--clean") clean; exit 0 ;;
    "-d") BMRB_DIR=$2; shift ;;
    "-fd") FASTA_SEQ_DIR=$2; shift ;;
    "-h"|"--help") print_help; exit 0 ;;
    "-m") MIN_PROT_LEN=$2; shift ;;
    "-v"|"--version") echo "This is `basename $0` release number $VERSION"; exit 0 ;;
    *) echo "Error: unknowk tag '$1'"; print_help; exit 1 ;;
  esac
  shift
done

\rm -f $LOG_FILE

################################################################################
#
################################################################################
#
# Check some variables
#
flag=0
if test ! -d $BMRB_DIR; then
  echo "**** Directory '$BMRB_DIR' does not exist"
  flag=1
fi

if test ! -d $FASTA_SEQ_DIR; then
  echo "**** Directory '$FASTA_SEQ_DIR' does not exist"
  flag=1
fi

if test $flag = "1"; then
  echo "**** Errors occured ****"
  print_help
  exit 1
fi

################################################################################
#
################################################################################
#
# Select proteins in BMRB files
#
echo " --> Selecting BMRB protein files"
echo "     (`\ls ${BMRB_DIR}/*.str | wc -l` star files in '${BMRB_DIR}' directory)"
if test ! -f $BMRB_PROT_FILES; then
  cd $BMRB_DIR
  for file in `\ls *.str`; do
    if test -n "`grep _Mol_polymer_class $file | grep protein`"; then
      echo ${BMRB_DIR}/$file >> $BMRB_PROT_FILES
    fi
  done
  cd $WORK_DIR
  \mv -f ${BMRB_DIR}/$BMRB_PROT_FILES .
else
  echo "   + File '$BMRB_PROT_FILES' already exists, using it"
fi
echo "     (`\cat $BMRB_PROT_FILES | wc -l` protein files)"

################################################################################
#
################################################################################
#
# Remove paramagnetic proteins
# The idea is to remove proteins with more than 5% chemical shifts
# at more than 5 sigma from the standard distributions
#
echo " --> Removing paramagnetic proteins"
if test ! -f $BMRB_NOMAG_PROT_FILES; then
  for file in `\cat $BMRB_PROT_FILES`; do
    echo "$file" >> $LOG_FILE
    num_cs=`BMRB_to_cs -n < $file 2>> $LOG_FILE`
    num_cs_filtered=`BMRB_to_cs -n -f -mf 5.0 < $file 2>> $LOG_FILE`
    if test "$num_cs" = "0"; then
      echo "Warning: num_cs = 0" >> $LOG_FILE
      num_cs=1
    fi
# Uncomment the test you want (and comment the other) :
# 5% of chemical shifts
    if test `echo "100.0 * ($num_cs - $num_cs_filtered) / $num_cs >= 5.0" | bc -l` == "0"; then
# 5 chemical shifts
#    if test `echo "($num_cs - $num_cs_filtered) >= 5" | bc` == "0"; then
      echo $file >> $BMRB_NOMAG_PROT_FILES
    fi
  done
else
  echo "   + File '$BMRB_NOMAG_PROT_FILES' already exists, using it"
fi
echo "     (`\cat $BMRB_NOMAG_PROT_FILES | wc -l` files remain)"

################################################################################
#
################################################################################
#
# Extract sequences
#
echo " --> Extracting sequences from BMRB files"
\rm -f $FASTA_PROT_FILES
for file in `cat $BMRB_NOMAG_PROT_FILES`; do
  tmp=${file##*/}
  out_file="$FASTA_SEQ_DIR/${tmp%.str}.fasta"
  echo "$out_file" >> $FASTA_PROT_FILES
  if test ! -f $out_file; then
    BMRB_to_fasta < $file > $out_file
  else
    echo "   + File '$out_file' already exists, using it" >> $LOG_FILE
  fi
done
echo "     (`\cat $FASTA_PROT_FILES | wc -l` sequences)"

################################################################################
#
################################################################################
#
# Remove too small sequences
#
echo " --> Removing sequences with less than $MIN_PROT_LEN residues"
if test ! -f $FASTA_FILTERED_FILES && test ! -f $PROT_NAME_FILE; then
  \rm -f $FASTA_FILTERED_FILES $PROT_NAME_FILE
  for file in `cat $FASTA_PROT_FILES`; do
    prot_size=`fasta_size < $file`
    if test $prot_size -ge $MIN_PROT_LEN; then
      echo $file >> $FASTA_FILTERED_FILES
      echo "`basename ${file%.fasta}`" >> $PROT_NAME_FILE
    fi
  done
else
  echo "   + files '$FASTA_FILTERED_FILES' and '$PROT_NAME_FILE' already exist, using them"
fi
echo "     (`\cat $FASTA_FILTERED_FILES | wc -l` sequences left)"

################################################################################
#
################################################################################
#
# Align all vs all with FASTA
#
echo " --> Aligning all vs all BMRB selected sequences"
align_all_vs_all.sh -fl $FASTA_FILTERED_FILES

################################################################################
#
################################################################################
#
# Calculate groups
#
echo " --> Building distance matrix"
if test ! -f $DIST_MAT; then
  cd fasta_results
  for file in `\ls *.out`; do
    fasta_res_to_eval -l ${WORK_DIR}/$PROT_NAME_FILE < $file >> $DIST_MAT 2>> $LOG_FILE
  done
  cd $WORK_DIR
  \mv -f fasta_results/$DIST_MAT .
else
  echo "   + File '$DIST_MAT' already exists, using it"
fi
echo "     (matrix has `\cat $DIST_MAT | wc -l` lines)"

#

echo " --> Calculating groups of homologous sequences"
if test ! -f $GROUPS_FILE; then
  find_min_groups -c 1e-3 < $DIST_MAT > $GROUPS_FILE 2>> $LOG_FILE
else
  echo "   + File '$GROUPS_FILE' already exists, using it"
fi
echo "     (`\cat $GROUPS_FILE | wc -l` groups were found)"

################################################################################
#
################################################################################
#
# Select the "best" sequence in each group
#
echo " --> Counting the number of chemical shifts for each sequence"
echo "     (pH min = 3, pH max = 10, T min = 283K, T max = 318K)"
if test ! -f $COUNT_CS_FILE; then
  num=0
  for name in `cat $PROT_NAME_FILE`; do
    echo "$name :" >> $LOG_FILE
    num=`echo "$num + 1" | bc`
    num_cs=`BMRB_to_cs -n -f -pH_min 3 -pH_max 10 -T_min 283 -T_max 318 < ${BMRB_DIR}/${name}.str 2>> $LOG_FILE`
    echo "num_cs = $num_cs" >> $LOG_FILE
    echo "num: $num, name: $name, num_cs: $num_cs" >> $COUNT_CS_FILE
  done
else
  echo "   + File '$COUNT_CS_FILE' already exists, using it"
fi
echo "     (`\cat $COUNT_CS_FILE | wc -l` sequences to choose from)"

echo " --> Selecting sequence with maximum number of chemical shifts for each group"
if test ! -f $FINAL_PROT_FILE; then
  for group in `cat $GROUPS_FILE | cut -d ':' -f 1 | tr ':' '\n'`; do
    \rm -f file.tmp
    for prot in `head -n $group $GROUPS_FILE | tail -n 1 | cut -c 7-`; do
      grep "num: $prot," $COUNT_CS_FILE >> file.tmp
    done
    name=`cat file.tmp | sort -nk 6 | tail -n 1 | cut -d ':' -f 3 | cut -d ',' -f 1`
    \rm -f file.tmp
    echo "      $name" >> $FINAL_PROT_FILE
  done
else
  echo "   + File '$FINAL_PROT_FILE' already exists, using it"
fi
echo "     (`\cat $FINAL_PROT_FILE | wc -l` sequences [1 sequence / group])"
