/*
 * This header is used by various programs i wrote.
 * Copyright (C) 2003 Antoine MARIN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1 published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact the author via mail or email at:
 * Antoine Marin (Antoine.Marin@ibpc.fr)
 * CNRS UPR 9080
 * Laboratoire de Biochimie Thorique
 * 13, rue Pierre et Marie Curie
 * 75005 Paris - France
 */

#define S(x) #x

#define MIN2(a, b)     ((a) < (b) ? (a) : (b))
#define MIN3(a, b, c)  ((a) < (b) ? MIN2((a), (c)) : MIN2((b), (c)))
#define MAX2(a, b)     ((a) > (b) ? (a) : (b))
#define MAX3(a, b, c)  ((a) > (b) ? MAX2((a), (c)) : MAX2((b), (c)))

#define PRINT_TAG_OUT { fprintf(stdout, "%s: at line %d in %s\n", \
    __FILE__, __LINE__, __FUNCTION__); fflush(stdout); }

#define PRINT_TAG_ERR { fprintf(stderr, "%s: at line %d in %s\n",\
    __FILE__, __LINE__, __FUNCTION__); fflush(stderr); }

/*! \def MALLOC
 * Memory allocation with some checkings.
 * For instance, the following code:
 *
 * int *toto;
 * int nel = 10;
 * MALLOC(toto, nel, int);
 *
 * is equivalent to:
 *
 * int *toto;
 * int nel = 10;
 * toto = (int *) malloc((size_t) nel * sizeof(int));
 */
#define MALLOC(ptr, nelem, type) {\
    (ptr) = (type *) malloc((size_t) (nelem) * sizeof(type));\
    if ((ptr) == NULL) {\
      PRINT_TAG_ERR\
      fprintf(stderr, "Unable to allocate memory (%d bytes) for %s\n",\
          (nelem) * sizeof(type), S(ptr));\
      exit(1);\
    }\
  }

#define WARN_MEM_SIZE 5000

#define CHK_ALLOC(ptr, nelem, type) {\
    if ((nelem) * sizeof(type) > WARN_MEM_SIZE) {\
      PRINT_TAG_OUT\
      fprintf(stdout, "On the way to allocate %d bytes for '%s' !\n",\
          (nelem) * sizeof(type), S(ptr));\
    }\
    (ptr) = (type *) malloc((size_t) (nelem) * sizeof(type));\
    if ((ptr) == NULL) {\
      PRINT_TAG_ERR\
      fprintf(stderr, "Unable to allocate memory (%d bytes) for %s\n",\
          (nelem) * sizeof(type), S(ptr));\
      exit(1);\
    }\
  }

/*! \def ALLOC
 * Define ALLOC to CHK_ALLOC to print when more than a given
 * memory size (defined by WARN_MEM_SIZE) is allocated,
 * else define it as MALLOC
 */
#define ALLOC MALLOC

/*! \def ALLOC2
 * Allocate a two dimensions array.
 * It only uses two malloc calls.
 */
#define ALLOC2(ptr, nelem1, nelem2, type) {\
    int alloc2_i;\
    ALLOC((ptr), (nelem1), type *);\
    ALLOC((ptr)[0], (nelem1) * (nelem2), type);\
    for (alloc2_i = 1; alloc2_i < (nelem1); alloc2_i++) {\
      (ptr)[alloc2_i] = (ptr)[alloc2_i - 1] + (nelem2);\
    }\
  }

/*! \def ALLOC2
 * Allocate a three dimensions array
 * It only uses three malloc calls.
 */
#define ALLOC3(ptr, nelem1, nelem2, nelem3, type) {\
    int alloc3_i;\
    ALLOC2((ptr), (nelem1), (nelem2), type *);\
    ALLOC((ptr)[0][0], (nelem1) * (nelem2) * (nelem3), type);\
    for (alloc3_i = 1; alloc3_i < ((nelem1) * (nelem2)); alloc3_i++) {\
      (ptr)[0][alloc3_i] = (ptr)[0][alloc3_i - 1] + (nelem3);\
    }\
  }

/*! \def CALLOC
 * Same as ALLOC but for CALLOC
 */
#define CALLOC(ptr, nelem, type) {\
    (ptr) = (type *) calloc((size_t) (nelem), (size_t) sizeof(type));\
    if ((ptr) == NULL) {\
      PRINT_TAG_ERR;\
      fprintf(stderr, "Unable to allocate memory (%d bytes) for %s\n",\
          (nelem) * sizeof(type), S(ptr));\
      exit(1);\
    }\
  }

/*! \def CALLOC2
 * Same as ALLOC2 but for CALLOC
 */
#define CALLOC2(ptr, nelem1, nelem2, type) {\
    int calloc2_i;\
    CALLOC((ptr), (nelem1), type *);\
    CALLOC((ptr)[0], (nelem1) * (nelem2), type);\
    for (calloc2_i = 1; calloc2_i < (nelem1); calloc2_i++) {\
      (ptr)[calloc2_i] = (ptr)[calloc2_i - 1] + (nelem2);\
    }\
  }

/*! \def ALLOC3
 * Same as ALLOC3 but for CALLOC
 */
#define CALLOC3(ptr, nelem1, nelem2, nelem3, type) {\
    int calloc3_i;\
    CALLOC2((ptr), (nelem1), (nelem2), type *);\
    CALLOC((ptr)[0][0], (nelem1) * (nelem2) * (nelem3), type);\
    for (calloc3_i = 1; calloc3_i < ((nelem1) * (nelem2)); calloc3_i++) {\
      (ptr)[0][calloc3_i] = (ptr)[0][calloc3_i - 1] + (nelem3);\
    }\
  }

/*! \def FREE
 * Redefine free to include some checkings
 */
#define FREE(ptr) {\
    if ((ptr) == NULL) {\
      PRINT_TAG_OUT\
      fprintf(stdout, "Trying to free a NULL pointer: %s\n", S(ptr));\
    } else {\
      free(ptr);\
      (ptr) = NULL;\
    }\
  }

/*! \def FREE2
 * Free an array allocated with ALLOC2
 */
#define FREE2(ptr) {\
    FREE((ptr)[0]);\
    FREE(ptr);\
  }

/*! \def FREE3
 * Free an array allocated with ALLOC3
 */
#define FREE3(ptr) {\
    FREE((ptr)[0][0]);\
    FREE((ptr)[0]);\
    FREE(ptr);\
  }

/*! \def FOPEN
 * Redefine fopen to include error check
 */
#define FOPEN(fp, fich, mode) {\
    (fp) = fopen((fich), (mode));\
    if ((fp) == NULL) {\
      PRINT_TAG_ERR\
      fprintf(stderr, "Unable to open file '%s'\n", (fich));\
      exit(1);\
    }\
  }

/*! \def FCLOSE
 * Redefine fclose to include warning check
 */
#define FCLOSE(fp) {\
    if ((fp) == NULL) {\
      PRINT_TAG_OUT\
      fprintf(stdout, "Trying to close a NULL file pointer: %s\n", S(fp));\
    } else {\
      if (fclose(fp) != 0) {\
        PRINT_TAG_OUT\
        fprintf(stdout, "Unable to close file: %s\n", S(fp));\
      }\
      (fp) = NULL;\
    }\
  }
