/*
 * This progam extracts the sequence in FASTA format from BMRB files.
 *
 * Copyright (C) 2003 Antoine MARIN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1 published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact the author via mail or email at:
 * Antoine Marin (Antoine.Marin@ibpc.fr)
 * CNRS UPR 9080
 * Laboratoire de Biochimie Thorique
 * 13, rue Pierre et Marie Curie
 * 75005 Paris - France
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>

#define VERSION "1.0"

#define MAX_BUF 1024
#define MAX_RES 10000
#define DEFAULT_OUTPUT_WIDTH 80

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
typedef struct parameters {
  int output_width;
} Parameters;

typedef struct data {
  int BMRB_acc_num;
  char desc[MAX_BUF];
  int seq_len;
  char seq[MAX_RES];
} Data;

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void init_parameters(Parameters *params)
{
  params->output_width = DEFAULT_OUTPUT_WIDTH;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_usage(FILE *fp, char **argv)
{
  fprintf(fp, "\nUsage: %s [-w width] < BMRB_file\n\n", basename(argv[0]));
  fprintf(fp, "  Option:\n\n");
  fprintf(fp, "    -h,--help     : print this help.\n\n");
  fprintf(fp, "    -v, --version : print version only (current version = %s).\n\n",
      VERSION);
  fprintf(fp, "    -w width      : integer value for the output width\n");
  fprintf(fp, "                    (default: %d), range: [%d - %d].\n\n",
      DEFAULT_OUTPUT_WIDTH, 1, MAX_RES);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void process_command_line(int argc, char **argv, Parameters *params)
{
  int i;

  for (i = 1; i < argc; i++) {

    if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {

/* -h, --help */

      print_usage(stdout, argv);
      exit(0);
    } else if (strcmp(argv[i], "-v") == 0 ||
      strcmp(argv[i], "--version") == 0) {

/* -v, --version : print version and exit */

      fprintf(stdout, "This is %s release number %s\n", basename(argv[0]),
          VERSION);
      exit(0);
    } else if (strcmp(argv[i], "-w") == 0) {

/* -w : output width */

      i++;
      params->output_width = atoi(argv[i]);
      if (params->output_width <= 0 || params->output_width > MAX_RES) {
        fprintf(stderr, "Error: width value out of range: '%d'\n",
            params->output_width);
        print_usage(stderr, argv);
        exit(1);
      }
    } else {

/* Unknown tag */

      fprintf(stderr, "Error: unknown tag: '%s'\n", argv[i]);
      print_usage(stderr, argv);
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int get_title(Data *data)
{
  char buffer[MAX_BUF], word[MAX_BUF], *sp;
  int tmp_len;

  while (fgets(buffer, MAX_BUF, stdin) != NULL) {
    if (sscanf(buffer, "%s", word) == 1) {
      if (strcmp(word, "_BMRB_accession_number") == 0) {

/* Get accession number */

        if (sscanf(buffer, "%s%d", word, &(data->BMRB_acc_num)) != 2) {
          fprintf(stderr, "Error: unable to find BMRB accession number\n");
          exit(1);
        }
      } else if (strcmp(word, "_Mol_system_name") == 0) {

/* Get molecule description */

        sp = strstr(buffer, "_Mol_system_name");
        sp += (char) 16;
        while (sp[0] == ' ') {
          sp++;
        }
        if (sp[0] != '\n') {
          tmp_len = strlen(sp);
          if (sp[tmp_len - 1] == '\n') {
            sp[tmp_len - 1] = '\0';
          }
          strcpy(data->desc, sp);
        } else {
          fprintf(stderr, "Warning: unable to find BMRB system name for "
              "file bmr%d.str\n", data->BMRB_acc_num);
        }
      } else if (strcmp(word, "_Residue_count") == 0) {

/* Get sequence length */

        if (sscanf(buffer, "%s%d", word, &(data->seq_len)) == 2) {
          return 0;
        } else {
          fprintf(stderr, "Warning: unable to find BMRB residue count\n");
          return 1;
        }
      }
    }
  }

  return 1;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int get_fasta_seq(Data *data)
{
  char buffer[MAX_BUF], word[128];
  int tmp_len;

  data->seq[0] = '\0';
  while (fgets(buffer, MAX_BUF, stdin) != NULL) {
    if (sscanf(buffer, "%s", word) == 1 &&
        strcmp(word, "_Mol_residue_sequence") == 0) {
      fgets(buffer, MAX_BUF, stdin);
      if (buffer[0] == ';') {
        fgets(buffer, MAX_BUF, stdin);
        while (buffer[0] != ';') {
          tmp_len = strlen(buffer);
          if (buffer[tmp_len - 1] == '\n') {
            buffer[tmp_len - 1] = '\0';
          }
          strcat(data->seq, buffer);
          if (fgets(buffer, MAX_BUF, stdin) == NULL) {
            return 1;
          }
        }
        if (data->seq[0] != '\0') {
          return 0;
        }
      } else {
        return 1;
      }
    }
  }

  return 1;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int check_fasta_seq(Data *data)
{
  int i, obs_seq_len;

  obs_seq_len = strlen(data->seq);
  for (i = 0; i < obs_seq_len; i++) {
    if (data->seq[i] < 'A' || data->seq[i] > 'Z') {
      obs_seq_len--;
      strcpy(data->seq + i, data->seq + i + 1);
      i--; /* stay at the same position */
    }
  }
  if (obs_seq_len != data->seq_len) {
    return obs_seq_len;
  }
  return 0;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void display_seq(char *seq, int from_pos, int to_pos, int width)
{
  int i;

  for (i = from_pos; i <= to_pos; i++) {
    printf("%c", seq[i]);
    if ((i + 1) % width == 0) {
      printf("\n");
    }
  }
  if (i % width != 0) {
    printf("\n");
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int main(int argc, char **argv)
{
  Parameters params;
  Data data;
  int obs_seq_len;

  init_parameters(&params);
  process_command_line(argc, argv, &params);

  if (get_title(&data) != 0) {
    fprintf(stderr, "Warning while getting title, some information may "
        "be missing for protein '%d'.\n", data.BMRB_acc_num);
  }

  rewind(stdin);
  if (get_fasta_seq(&data) == 0) {
    obs_seq_len = check_fasta_seq(&data);
    if (obs_seq_len != 0) {
      fprintf(stderr, "Warning: observed sequence lenght (%d) is "
          "different from the one in the '_Residue_count' field (%d) "
          "for protein '%d'\n", obs_seq_len, data.seq_len, data.BMRB_acc_num);

/* change the value for the observed length */

      data.seq_len = obs_seq_len;
    }
    printf(">bmr%d molecule: %s length: %d\n", data.BMRB_acc_num,
        data.desc, data.seq_len);
    display_seq(data.seq, 0, data.seq_len - 1, params.output_width);
  } else {
    fprintf(stderr, "Error: unable to get sequence for file 'bmr%d.str'.\n",
        data.BMRB_acc_num);
    fprintf(stderr, "Please check your input file.\n");
    exit(1);
  }

  return 0;
}
